@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2014  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -30,36 +30,12 @@
 
 #include <ell/ell.h>
 
-#include "linux/nl80211.h"
-
 #include "src/agent.h"
 #include "src/iwd.h"
 #include "src/dbus.h"
 
 static struct l_dbus *g_dbus = NULL;
 
-const char *dbus_iftype_to_string(uint32_t iftype)
-{
-	switch (iftype) {
-	case NL80211_IFTYPE_ADHOC:
-		return "ad-hoc";
-	case NL80211_IFTYPE_STATION:
-		return "station";
-	case NL80211_IFTYPE_AP:
-		return "ap";
-	case NL80211_IFTYPE_P2P_CLIENT:
-		return "p2p-client";
-	case NL80211_IFTYPE_P2P_GO:
-		return "p2p-go";
-	case NL80211_IFTYPE_P2P_DEVICE:
-		return "p2p-device";
-	default:
-		break;
-	}
-
-	return NULL;
-}
-
 struct l_dbus_message *dbus_error_busy(struct l_dbus_message *msg)
 {
 	return l_dbus_message_new_error(msg, IWD_SERVICE ".InProgress",
@@ -86,7 +62,7 @@ struct l_dbus_message *dbus_error_not_available(struct l_dbus_message *msg)
 
 struct l_dbus_message *dbus_error_invalid_args(struct l_dbus_message *msg)
 {
-	return l_dbus_message_new_error(msg, IWD_SERVICE ".InvalidArgs",
+	return l_dbus_message_new_error(msg, IWD_SERVICE ".InvalidArguments",
 					"Argument type is wrong");
 }
 
@@ -202,6 +178,32 @@ void dbus_pending_reply(struct l_dbus_message **msg,
 	*msg = NULL;
 }
 
+/*
+ * Convenience helper for appending a dictionary "{sv}". This only works when
+ * the variant is a basic type.
+ */
+bool dbus_append_dict_basic(struct l_dbus_message_builder *builder,
+				const char *name, char type,
+				const void *data)
+{
+	char strtype[] = { type, '\0' };
+
+	if (!l_dbus_message_builder_enter_dict(builder, "sv"))
+		return false;
+	if (!l_dbus_message_builder_append_basic(builder, 's', name))
+		return false;
+	if (!l_dbus_message_builder_enter_variant(builder, strtype))
+		return false;
+	if (!l_dbus_message_builder_append_basic(builder, type, data))
+		return false;
+	if (!l_dbus_message_builder_leave_variant(builder))
+		return false;
+	if (!l_dbus_message_builder_leave_dict(builder))
+		return false;
+
+	return true;
+}
+
 struct l_dbus *dbus_get_bus(void)
 {
 	return g_dbus;
