@@ -114,6 +114,7 @@ enum hwsim_tx_control_flags {
 
 #define IEEE80211_TX_RATE_TABLE_SIZE	4
 #define HWSIM_DELAY_MIN_MS		1
+#define HWSIM_MAX_PREFIX_LEN		128
 
 struct hwsim_rule {
 	unsigned int id;
@@ -123,10 +124,13 @@ struct hwsim_rule {
 	bool destination_any : 1;
 	bool bidirectional : 1;
 	bool drop : 1;
+	bool enabled : 1;
 	uint32_t frequency;
 	int priority;
 	int signal;
 	int delay;
+	uint8_t *prefix;
+	size_t prefix_len;
 };
 
 struct hwsim_support {
@@ -1174,6 +1178,9 @@ static void process_rules(const struct radio_info_rec *src_radio,
 			rule_entry = rule_entry->next) {
 		struct hwsim_rule *rule = rule_entry->data;
 
+		if (!rule->enabled)
+			return;
+
 		if (!rule->source_any &&
 				!radio_match_addr(src_radio, rule->source) &&
 				(!rule->bidirectional ||
@@ -1203,6 +1210,12 @@ static void process_rules(const struct radio_info_rec *src_radio,
 		if (rule->frequency && rule->frequency != frame->frequency)
 			continue;
 
+		if (rule->prefix && frame->payload_len >= rule->prefix_len) {
+			if (memcmp(rule->prefix, frame->payload,
+					rule->prefix_len) != 0)
+				continue;
+		}
+
 		/* Rule deemed to match frame, apply any changes */
 
 		if (rule->signal)
@@ -1423,9 +1436,11 @@ static void frame_delay_callback(struct l_timeout *timeout, void *user_data)
 	else
 		send_frame_destroy(send_info);
 
-	l_timeout_remove(timeout);
+	if (timeout)
+		l_timeout_remove(timeout);
 }
 
+
 /*
  * Process frames in a similar way to how the kernel built-in hwsim medium
  * does this, with an additional optimization for unicast frames and
@@ -1444,7 +1459,7 @@ static void process_frame(struct hwsim_frame *frame)
 		struct radio_info_rec *radio = entry->data;
 		struct send_frame_info *send_info;
 		bool drop = drop_mcast;
-		uint32_t delay = HWSIM_DELAY_MIN_MS;
+		uint32_t delay = 0;
 
 		if (radio == frame->src_radio)
 			continue;
@@ -1486,11 +1501,16 @@ static void process_frame(struct hwsim_frame *frame)
 		send_info->radio = radio;
 		send_info->frame = hwsim_frame_ref(frame);
 
-		if (!l_timeout_create_ms(delay, frame_delay_callback,
-						send_info, NULL)) {
-			l_error("Error delaying frame, frame will be dropped");
-			send_frame_destroy(send_info);
-		}
+		if (delay) {
+			if (!l_timeout_create_ms(delay, frame_delay_callback,
+							send_info, NULL)) {
+				l_error("Error delaying frame %ums, "
+						"frame will be dropped", delay);
+				send_frame_destroy(send_info);
+			}
+		} else
+			frame_delay_callback(NULL, send_info);
+
 	}
 
 	hwsim_frame_unref(frame);
@@ -1688,8 +1708,7 @@ static struct l_dbus_message *radio_manager_create(struct l_dbus *dbus,
 			goto invalid;
 	}
 
-	new_msg = l_genl_msg_new_sized(HWSIM_CMD_NEW_RADIO,
-					16 + name ? strlen(name) : 0);
+	new_msg = l_genl_msg_new(HWSIM_CMD_NEW_RADIO);
 	l_genl_msg_append_attr(new_msg, HWSIM_ATTR_DESTROY_RADIO_ON_CLOSE,
 				0, NULL);
 
@@ -1987,7 +2006,8 @@ static struct l_dbus_message *rule_add(struct l_dbus *dbus,
 	rule->id = next_rule_id++;
 	rule->source_any = true;
 	rule->destination_any = true;
-	rule->delay = HWSIM_DELAY_MIN_MS;
+	rule->delay = 0;
+	rule->enabled = false;
 
 	if (!rules)
 		rules = l_queue_new();
@@ -2026,6 +2046,10 @@ static struct l_dbus_message *rule_remove(struct l_dbus *dbus,
 
 	path = rule_get_path(rule);
 	l_queue_remove(rules, rule);
+
+	if (rule->prefix)
+		l_free(rule->prefix);
+
 	l_free(rule);
 	l_dbus_unregister_object(dbus, path);
 
@@ -2304,6 +2328,90 @@ static struct l_dbus_message *rule_property_set_delay(
 	return l_dbus_message_new_method_return(message);
 }
 
+static bool rule_property_get_prefix(struct l_dbus *dbus,
+					struct l_dbus_message *message,
+					struct l_dbus_message_builder *builder,
+					void *user_data)
+{
+	struct hwsim_rule *rule = user_data;
+	size_t i;
+
+	l_dbus_message_builder_enter_array(builder, "y");
+
+	for (i = 0; i < rule->prefix_len; i++)
+		l_dbus_message_builder_append_basic(builder, 'y',
+							rule->prefix + i);
+
+	l_dbus_message_builder_leave_array(builder);
+
+	return true;
+}
+
+static struct l_dbus_message *rule_property_set_prefix(
+					struct l_dbus *dbus,
+					struct l_dbus_message *message,
+					struct l_dbus_message_iter *new_value,
+					l_dbus_property_complete_cb_t complete,
+					void *user_data)
+{
+	struct hwsim_rule *rule = user_data;
+	struct l_dbus_message_iter iter;
+	const uint8_t *prefix;
+	uint32_t len;
+
+	if (!l_dbus_message_iter_get_variant(new_value, "ay", &iter))
+		goto invalid_args;
+
+	if (!l_dbus_message_iter_get_fixed_array(&iter,
+						(const void **)&prefix, &len))
+		goto invalid_args;
+
+	if (len > HWSIM_MAX_PREFIX_LEN)
+		goto invalid_args;
+
+	if (rule->prefix)
+		l_free(rule->prefix);
+
+	rule->prefix = l_memdup(prefix, len);
+	rule->prefix_len = len;
+
+	return l_dbus_message_new_method_return(message);
+
+invalid_args:
+	return dbus_error_invalid_args(message);
+}
+
+static bool rule_property_get_enabled(struct l_dbus *dbus,
+					struct l_dbus_message *message,
+					struct l_dbus_message_builder *builder,
+					void *user_data)
+{
+	struct hwsim_rule *rule = user_data;
+	bool bval = rule->enabled;
+
+	l_dbus_message_builder_append_basic(builder, 'b', &bval);
+
+	return true;
+}
+
+static struct l_dbus_message *rule_property_set_enabled(
+					struct l_dbus *dbus,
+					struct l_dbus_message *message,
+					struct l_dbus_message_iter *new_value,
+					l_dbus_property_complete_cb_t complete,
+					void *user_data)
+{
+	struct hwsim_rule *rule = user_data;
+	bool bval;
+
+	if (!l_dbus_message_iter_get_variant(new_value, "b", &bval))
+		return dbus_error_invalid_args(message);
+
+	rule->enabled = bval;
+
+	return l_dbus_message_new_method_return(message);
+}
+
 static void setup_rule_interface(struct l_dbus_interface *interface)
 {
 	l_dbus_interface_method(interface, "Remove", 0, rule_remove, "", "");
@@ -2340,6 +2448,14 @@ static void setup_rule_interface(struct l_dbus_interface *interface)
 					L_DBUS_PROPERTY_FLAG_AUTO_EMIT, "u",
 					rule_property_get_delay,
 					rule_property_set_delay);
+	l_dbus_interface_property(interface, "Prefix",
+					L_DBUS_PROPERTY_FLAG_AUTO_EMIT, "ay",
+					rule_property_get_prefix,
+					rule_property_set_prefix);
+	l_dbus_interface_property(interface, "Enabled",
+					L_DBUS_PROPERTY_FLAG_AUTO_EMIT, "b",
+					rule_property_get_enabled,
+					rule_property_set_enabled);
 }
 
 static void request_name_callback(struct l_dbus *dbus, bool success,
